from __future__ import annotations

import abc
import enum
import logging
from io import StringIO
from typing import Optional, List, Any

import attr
from pg_purepy.exc import ProtocolParseError, PostgresqlError

logging.basicConfig(level=logging.DEBUG)


@attr.s(slots=True, frozen=True)
class PostgresMessage(object):
    """
    Base class for a PostgreSQL protocol message.
    """


class AuthenticationMethod(enum.IntEnum):
    """
    Enumeration of supported authentication methods.
    """
    #: The server wishes for us to send our password in clear text.
    CLEARTEXT = 3
    #: The server wishes for us to use MD5 hashing for our password.
    MD5 = 5
    #: The server wishes for us to use SASL authentication.
    SASL = 10


@attr.s(slots=True, frozen=True)
class AuthenticationRequest(PostgresMessage):
    """
    Returned when the PostgreSQL server requires authentication.
    """

    #: The authentication method required.
    method: AuthenticationMethod = attr.ib()

    #: When doing MD5 authentication, the salt to use.
    md5_salt: Optional[bytes] = attr.ib(default=None)

    #: When doing SASL authentication, the list of authentication methods.
    sasl_methods: List[str] = attr.ib(default=[])


@attr.s(slots=True, frozen=True)
class BackendKeyData(PostgresMessage):
    """
    Misc data used for cancellation.
    """

    #: The :class:`int` PID of this connection.
    pid: int = attr.ib()

    #: The 64-biit :class:`int` secret key data of this connection.
    secret_key: int = attr.ib()


@attr.s(slots=True, frozen=True)
class AuthenticationCompleted(PostgresMessage):
    """
    Returned when authentication is completed.
    """


@attr.s(slots=True, frozen=True)
class ParameterStatus(PostgresMessage):
    """
    Returned when a configuration parameter is changed, e.g. via SET.
    """

    #: The :class:`str` name of the parameter.
    name: str = attr.ib()

    #: The :class:`str` value of the parameter.
    value: str = attr.ib()


class ReadyForQueryState(enum.Enum):
    """
    Enumeration of possible ReadyForQuery states.
    """

    #: The server is idle and not in any transaction.
    IDLE = ord("I")

    #: The server is currently in a transaction, and new commands can be issued.
    IN_TRANSACTION = ord("T")

    #: The server is currently in a transaction that has errored, and no new commands can be issued.
    ERRORED_TRANSACTION = ord("E")


@attr.s(slots=True, frozen=True)
class ReadyForQuery(PostgresMessage):
    """
    Returned when the protocol machine is ready for the next query cycle.
    """

    #: The sub-state that the current connection is in.
    state: ReadyForQueryState = attr.ib()


class ErrorResponseFieldType(enum.IntEnum):
    """
    Enumeration of possible error response fields.
    """

    SEVERITY_LOCALISED = ord("S")
    SEVERITY = ord("V")
    CODE = ord("C")
    MESSAGE = ord("M")
    DETAIL = ord("D")
    HINT = ord("H")
    POSITION = ord("P")
    INTERNAL_POSITION = ord("p")
    INTERNAL_QUERY = ord("q")
    WHERE = ord("W")
    SCHEMA_NAME = ord("s")
    TABLE_NAME = ord("t")
    COLUMN_NAME = ord("c")
    DATA_TYPE_NAME = ord("t")
    CONSTRAINT_NAME = ord("n")

    # 0 will never exist
    UNKNOWN = 0


@attr.s(slots=True, frozen=True)
class ColumnDescription:
    """
    A description of a column.
    """

    #: The :class:`str` name of this column.
    name: str = attr.ib()

    #: The optional :class:`int` table OID of this column.
    table_oid: Optional[int] = attr.ib()

    #: The optional :class:`int` column index of this column.
    column_index: Optional[int] = attr.ib()

    #: The :class`int` type OID of this column.
    type_oid: int = attr.ib()

    #: The :class`int` internal column length.
    column_length: int = attr.ib()

    #: The :class:`int` type modifier for this column.
    type_modifier: int = attr.ib()


@attr.s(slots=True, frozen=True)
class QueryResultMessage(PostgresMessage, abc.ABC):
    """
    Superclass for query results.
    """


@attr.s(slots=True, frozen=True)
class RowDescription(QueryResultMessage):
    """
    Describes the rows of a query.
    """
    #: The list of :class:`.ColumnDescription` instances that wraps the decoding info for each
    #: column returned in this row.
    columns: List[ColumnDescription] = attr.ib()


@attr.s(slots=True, frozen=True)
class DataRow(QueryResultMessage):
    """
    A singular data row. This contains a :class:`.RowDescription` and a list of converted data
    values.
    """

    #: The :class:`.RowDescription` that describes the data within this row.
    description: RowDescription = attr.ib()

    #: A list of column values, in the same order as the description, that contains the actual
    #: converted data incoming from the server.
    data: List[Any] = attr.ib()


@attr.s(slots=True, frozen=True)
class CommandComplete(QueryResultMessage):
    """
    Returned when a single query command is complete.
    """

    #: The :class:`str` command tag. Probably useless.
    tag: str = attr.ib()

    #: The :class:`int` row count returned. This may be None if the command does not have a row
    #: count (e.g. SHOW or SET).
    row_count: Optional[int] = attr.ib()


@attr.s(slots=True, frozen=True)
class ParseComplete(PostgresMessage):
    """
    Returned when parsing a prepared statement completes.
    """

    #: The :class:`str` name of the statement prepared. None means the unnamed prepared statement.
    statement_name: Optional[str] = attr.ib()


@attr.s(slots=True, frozen=True)
class ParameterDescription(PostgresMessage):
    """
    Returned when parsing a ParameterDescription message.
    """

    #: The list of :class:`int` OIDs within this description.
    oids: List[int] = attr.ib()


@attr.s(slots=True, frozen=True)
class PreparedStatementInfo(PostgresMessage):
    """
    Contains the state of a prepared statement. Returned for a RowDescription over a prepared
    statement.
    """

    #: The :class:`str` name of the prepared statement.
    name: Optional[str] = attr.ib()

    #: The :class:`~.ParameterDescription` for the parameters for this prepared statement.
    parameter_oids: ParameterDescription = attr.ib()

    #: The :class:`~.RowDescription` of the incoming row data of the prepared statement.
    #: This may be None if this query doesn't return any data.
    row_description: Optional[RowDescription] = attr.ib()


@attr.s(slots=True, frozen=True)
class BindComplete(PostgresMessage):
    """
    Returned when a Bind message completes successfully.
    """


def _optional_int(value: Optional[str]) -> Optional[int]:
    if value is None:
        return None

    return int(value)


@attr.s(slots=True, frozen=True)
class ErrorOrNoticeResponse(PostgresMessage):
    """
    Returned when an error or a notice message is produced from the server.
    """

    #: If this error is a notice, rather than a real error.
    notice: bool = attr.ib()

    #: If this error is recoverable or not.
    recoverable: bool = attr.ib()

    severity_localised: str = attr.ib()
    severity: str = attr.ib()
    code: str = attr.ib()
    message: str = attr.ib()

    # optional parameters
    detail: Optional[str] = attr.ib(default=None)
    position: Optional[int] = attr.ib(default=None, converter=_optional_int)
    internal_position: Optional[int] = attr.ib(default=None, converter=_optional_int)
    internal_query: Optional[str] = attr.ib(default=None)
    where: Optional[str] = attr.ib(default=None)
    schema_name: Optional[str] = attr.ib(default=None)
    table_name: Optional[str] = attr.ib(default=None)
    column_name: Optional[str] = attr.ib(default=None)
    data_type_name: Optional[str] = attr.ib(default=None)
    constraint_name: Optional[str] = attr.ib(default=None)


class BaseDatabaseError(PostgresqlError):
    """
    An exception produceed from the database, usually from an ErrorOrNoticeResponse message. This does
    NOT include things such as protocol parsing errors.
    """

    def __init__(self, response: ErrorOrNoticeResponse, query: str = None):
        self.response = response

    def __str__(self) -> str:
        buf = StringIO()
        buf.write(self.response.severity)
        buf.write(":")
        buf.write(" [code ")
        buf.write(str(self.response.code))
        buf.write("] ")
        buf.write(self.response.message)

        # TODO: Optional fields
        return buf.getvalue()


class RecoverableDatabaseError(BaseDatabaseError):
    """
    A subclass of :class:`.BaseDatabaseError` that the client may potentially recover from. Examples
    include query errors.
    """


class UnrecoverableDatabaseError(BaseDatabaseError):
    """
    A subclass of :class:`.BaseDatabaseError` that the client must not recover from. This usually
    implies internal errors in the server.
    """


class InvalidPasswordError(UnrecoverableDatabaseError):
    """
    Raised when the password provided is invalid.
    """


def wrap_error(response: ErrorOrNoticeResponse, query: str = None) -> BaseDatabaseError:
    """
    Wraps a :class:`.ErrorOrNoticeResponse` in an exception. If a query produced the error in question, then
    passing it as the ``query`` param can produce a prettier error.
    """
    # TODO: More codes
    if response.code == "28P01":
        return InvalidPasswordError(response, query)

    else:
        if response.recoverable:
            return RecoverableDatabaseError(response, query)
        else:
            return UnrecoverableDatabaseError(response, query)


__all__ = (
    "PostgresMessage",
    "AuthenticationRequest",
    "BackendKeyData",
    "AuthenticationCompleted",
    "ParameterStatus",
    "ReadyForQuery",
    "QueryResultMessage",
    "RowDescription",
    "DataRow",
    "CommandComplete",
    "ParseComplete",
    "ParameterDescription",
    "PreparedStatementInfo",
    "BindComplete",
    "ErrorOrNoticeResponse",

    "InvalidPasswordError",
    "BaseDatabaseError",
    "RecoverableDatabaseError",
    "UnrecoverableDatabaseError",
)



from collections import deque
from typing import List


def pack_strings(*s: str, encoding: str = "ascii") -> bytes:
    """
    Packs a sequence of strings using null terminators.
    """
    return b"\x00".join(x.encode(encoding) for x in s) + b"\x00"


class Buffer(object):
    """
    Simple buffer that allows reading data off of a bytearray ala Java ByteBuffer.
    """

    def __init__(self, ba: bytearray = None):
        self.data = deque(ba if ba else b"")

    def __bool__(self):
        return bool(self.data)

    def __len__(self):
        return len(self.data)

    def read_bytes(self, count: int):
        ba = bytearray()
        for x in range(0, count):
            ba.append(self.data.popleft())

        return ba

    def read_byte(self) -> int:
        return self.data.popleft()

    def read_short(self) -> int:
        return int.from_bytes(self.read_bytes(2), byteorder="big", signed=True)

    def read_int(self) -> int:
        return int.from_bytes(self.read_bytes(4), byteorder="big", signed=True)

    def read_long(self) -> int:
        return int.from_bytes(self.read_bytes(8), byteorder="big", signed=True)

    def read_cstring(self, encoding: str) -> str:
        """
        Reads a null-terminated C string.
        """
        buf = bytearray()
        while True:
            byte = self.data.popleft()
            if byte == 0x0:
                break

            buf.append(byte)

        return buf.decode(encoding=encoding)

    def read_all_cstrings(self, encoding: str, drop_empty: bool = False) -> List[str]:
        """
        Reads all null-terminated C strings from the buffer.
        """
        items = []
        while self.data:
            items += self.read_cstring(encoding)

        if drop_empty:
            return [i for i in items if not i]
        else:
            return items



"""
Index of builtin converters.
"""
import abc
from typing import Any, Callable, Tuple, Dict, List

import attr


@attr.s(slots=True, frozen=False)
class ConversionContext:
    """
    A conversion context contains information that might be needed to convert from the PostgreSQL
    string representation to the real representation.
    """

    #: The encoding of the client.
    client_encoding: str = attr.ib()


class Converter(abc.ABC):
    """
    A converter converts between the PostgreSQL representation of data and the Python
    representation, in either direction.
    """

    #: The OID of the PostgreSQL type this converter uses.
    oid: int

    @abc.abstractmethod
    def from_postgres(self, context: ConversionContext, data: str) -> Any:
        """
        Converts ``data`` from the PostgreSQL string representation to a Python type.
        """

    @abc.abstractmethod
    def to_postgres(self, context: ConversionContext, data: Any) -> str:
        """
        Converts ``data`` from the Python type to the PostgreSQL string representation.
        """


class SimpleFunctionConverter(Converter):
    """
    A simple converter that calls fn_from_pg to convert to a Python type, and fn_to_pg to convert
    from a Python type.
    """

    def __init__(self, oid: int, fn_from_pg, fn_to_pg):
        self.oid = oid

        self._to = fn_from_pg
        self._from = fn_to_pg

    def from_postgres(self, context: ConversionContext, data: str) -> Any:
        return self._to(data)

    def to_postgres(self, context: ConversionContext, data: Any) -> str:
        return self._from(data)


class BoolConverter(Converter):
    """
    Converter that converts for booleans.
    """

    def __init__(self):
        self.oid = 16

    def from_postgres(self, context: ConversionContext, data: str) -> Any:
        return data == "t"

    def to_postgres(self, context: ConversionContext, data: Any) -> str:
        return "true" if data else "false"


class TextConverter(Converter):
    """
    Pass-through converter for the various text types.
    """

    def __init__(self, oid: int):
        self.oid = oid

    def from_postgres(self, context: ConversionContext, data: str) -> Any:
        return data

    def to_postgres(self, context: ConversionContext, data: Any) -> str:
        return data


_PASSTHROUGH = lambda it: it
DEFAULT_TEXT_CONVERTER = SimpleFunctionConverter(-1, _PASSTHROUGH, _PASSTHROUGH)

#: A list of the default type converters.
DEFAULT_CONVERTERS: List[Converter] = [
    BoolConverter(),  # bool
    SimpleFunctionConverter(20, int, str),  # int8
    SimpleFunctionConverter(21, int, str),  # int2
    SimpleFunctionConverter(23, int, str),  # int4
    SimpleFunctionConverter(26, int, str),  # oid
    SimpleFunctionConverter(18, _PASSTHROUGH, _PASSTHROUGH),  # char
    SimpleFunctionConverter(19, _PASSTHROUGH, _PASSTHROUGH),  # name
    SimpleFunctionConverter(25, _PASSTHROUGH, _PASSTHROUGH),  # text
]



class PostgresqlError(Exception):
    """
    Base exception class all other exceptions are derived from.
    """


class ProtocolParseError(PostgresqlError):
    """
    Base exception class for all protocol parsing related errors.
    """


class MissingPasswordError(ProtocolParseError):
    """
    Raised when the server asks for a password, but we don't have one.
    """


class UnknownMessageError(ProtocolParseError):
    """
    Raised when an unknown message is returned.
    """


class IllegalStateError(ProtocolParseError):
    """
    Raised when an operation is attempted that would result in an illegal state.
    """



from pg_purepy.connection import open_database_connection, AsyncPostgresConnection, QueryResult
from pg_purepy.exc import *
from pg_purepy.messages import *
from pg_purepy.protocol import (
    SansIOClient,
    ProtocolState,
    NEED_DATA,
    SSL_MESSAGE,
    check_if_tls_accepted,
)



"""
The AnyIO implementation of the PostgreSQL client.
"""

from __future__ import annotations

import warnings
from contextlib import asynccontextmanager
from os import PathLike
from ssl import SSLContext
from typing import (
    Union,
    AsyncContextManager,
    AsyncIterator,
    TypeVar,
    Type,
    List,
    Optional,
)

import anyio
from anyio import Lock
from anyio.abc import ByteStream
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from anyio.streams.tls import TLSStream
from pg_purepy.dbapi import convert_paramstyle
from pg_purepy.exc import IllegalStateError
from pg_purepy.messages import (
    ErrorOrNoticeResponse,
    wrap_error,
    QueryResultMessage,
    PreparedStatementInfo,
    BaseDatabaseError,
    BindComplete,
    PostgresMessage,
    DataRow,
    CommandComplete,
)
from pg_purepy.protocol import (
    SansIOClient,
    SSL_MESSAGE,
    check_if_tls_accepted,
    ProtocolParseError,
    NEED_DATA,
    ReadyForQuery,
)

try:
    from contextlib import aclosing
except ImportError:
    from async_generator import aclosing

T = TypeVar("T")


class AsyncPostgresConnection(object):
    """
    An asynchronous connection to a PostgreSQL server. This class should not be directly
    instantiated; instead, use :meth:`.open_database_connection`.
    """

    def __init__(
        self,
        stream: ByteStream,
        state: SansIOClient,
    ):
        self._stream = stream
        self._protocol = state

        self._query_lock = Lock()

    @property
    def ready(self) -> bool:
        """
        Returns if this connection is ready for another query.
        """
        return self._protocol.ready

    @property
    def in_transaction(self) -> bool:
        """
        Returns if this connection is currently in a transaction.
        """
        return self._protocol.in_transaction

    async def _do_startup(self):
        """
        Sends the startup message.
        """
        data = self._protocol.do_startup()
        await self._stream.send(data)

    async def _terminate(self):
        """
        Terminates the protocol. Does not close the connection.
        """
        data = self._protocol.do_terminate()
        await self._stream.send(data)

    async def _read_until_ready(self):
        """
        Yields events until the connection is ready. This is an asynchronous generator. You can
        discard events you don't care about.

        This must ONLY be called if the protocol is NOT ready.
        """
        if self._protocol.ready:
            await anyio.sleep(0)
            return

        while True:
            while True:
                next_event = self._protocol.next_event()
                if next_event is NEED_DATA:
                    break

                if isinstance(next_event, ErrorOrNoticeResponse) and next_event.notice:
                    if next_event.severity == "WARNING":
                        err = wrap_error(next_event)
                        warnings.warn(str(err))

                yield next_event

                if isinstance(next_event, ReadyForQuery):
                    await anyio.sleep(0)  # checkpoint()
                    return

            to_send = self._protocol.get_needed_synchronisation()
            if to_send:
                await self._stream.send(to_send)

            received = await self._stream.receive(65536)
            self._protocol.receive_bytes(received)

    async def wait_until_ready(self):
        """
        Waits until the conneection is ready. This discards all events. Useful in the authentication
        loop.
        """
        async with aclosing(self._read_until_ready()) as gen:
            async for message in gen:
                if isinstance(message, ErrorOrNoticeResponse):
                    err = wrap_error(message)
                    raise err

    async def _wait_for_message(self, typ: Type[T], *, wait_until_ready: bool = True) -> T:
        """
        Waits until a message of type ``typ`` arrives.

        This will wait until the ReadyForQuery message arrives to avoid requiring extra
        synchronisation, if ``wait_until_ready`` is True. If it never arrives, this will deadlock!
        """
        message_found = None
        async with aclosing(self._read_until_ready()) as gen:
            async for item in gen:
                if isinstance(item, typ):
                    if not wait_until_ready:
                        return item

                    message_found = item

                elif isinstance(item, ErrorOrNoticeResponse):
                    raise wrap_error(item)

            if message_found is None:
                raise IllegalStateError(f"No message of type {typ} was yielded")

            return message_found

    ## Low-level API ##
    async def create_prepared_statement(self, name: str, query: str) -> PreparedStatementInfo:
        """
        Creates a prepared statement. This is part of the low-level query API.

        :param name: The name of the prepared statement.
        :param query: The query to use.
        """
        to_send = self._protocol.do_create_prepared_statement(name=name, query_text=query)
        await self._stream.send(to_send)
        pc = await self._wait_for_message(PreparedStatementInfo)
        return pc

    async def lowlevel_query(
        self,
        query: Union[str, PreparedStatementInfo],
        *params,
        **kwargs,
    ) -> AsyncIterator[QueryResultMessage]:
        """
        Performs a query to the server. This is an asynchronous generator; you must iterate over
        values in order to get the messages returned from the server.
        """
        async with self._query_lock:
            # always wait until ready! we do not like getting random messages from the last client
            # intermixed
            if not self._protocol.ready:
                await self.wait_until_ready()

            simple_query = not ((params or kwargs) or isinstance(query, PreparedStatementInfo))
            if simple_query:
                data = self._protocol.do_simple_query(query)
                await self._stream.send(data)
            else:
                if not isinstance(query, PreparedStatementInfo):
                    real_query, params = convert_paramstyle(query, kwargs)
                    info = await self.create_prepared_statement(name="", query=real_query)
                else:
                    info = query

                bound_data = self._protocol.do_bind_execute(info, params)
                await self._stream.send(bound_data)
                # we need to get BindComplete because we need to yield the statement's
                # RowDescription out, for a more "consistent" view.
                await self._wait_for_message(BindComplete, wait_until_ready=False)
                # no error, so the query is gonna complete successfully
                yield info.row_description

            async with aclosing(self._read_until_ready()) as agen:
                async for message in agen:
                    if isinstance(message, ErrorOrNoticeResponse) and not message.notice:
                        err = wrap_error(message)
                        raise err

                    if isinstance(message, QueryResultMessage):
                        yield message

    ## Mid-level API. ##
    @asynccontextmanager
    async def query(
        self,
        query: Union[str, PreparedStatementInfo],
        *params,
        **kwargs,
    ) -> AsyncContextManager[QueryResult]:
        """
        Mid-level query API.

        The ``query`` parameter can either be a string or a :class:`~.PreparedStatementInfo`, as
        returned from :func:`~.create_prepared_statement`. If it is a string, and it has parameters,
        they must be provided as keyword arguments. If it is a pre-prepared statement, and it has
        parameters, they must be provided as positional arguments.

        If keyword arguments are provided or a prepared statement is passed, an extended query with
        secure argument parsing will be used. Otherwise, a simple query will be used, which saves
        bandwidth over the extended query protocol.

        If the server is currently in a failed transaction, then your query will be ignored. Make
        sure to issue a rollback beforehand, if needed.

        This is an asynchronous context manager that yields a :class:`.QueryResult`, that can
        be asynchronously iterated over for the data rows of the query. Once all data rows have
        been iterated over, you can call :func:`.QueryResult.row_count` to get the total row count.
        """
        async with aclosing(self.lowlevel_query(query, *params, **kwargs)) as agen:
            yield QueryResult(agen.__aiter__())
            # always wait
            await self.wait_until_ready()

    @asynccontextmanager
    async def with_transaction(self) -> None:
        """
        Asynchronous context manager that automatically opens and closes a transaction.
        """

        try:
            await self.execute("begin;")
            yield
        except Exception:
            await self.execute("rollback;")
            raise
        else:
            await self.execute("commit;")

    ### DBAPI style methods ###
    async def fetch(
        self, query: Union[str, PreparedStatementInfo], *params, **kwargs
    ) -> List[DataRow]:
        """
        Eagerly fetches the result of a query. This returns a list of :class:`~.DataRow` objects.

        If you wish to lazily load the results of a query, use
        :meth:`~.AsyncPostgresConnection.query` instead.

        :param query: Either a :class:`str` that contains the query text,
                      or a :class:`~.PreparedStatementInfo` that represents a pre-prepared query.
        """
        async with self.query(query, *params, **kwargs) as q:
            return [i async for i in q]

    async def execute(self, query: Union[str, PreparedStatementInfo], *params, **kwargs) -> int:
        """
        Executes a query, returning its row count. This will discard all data rows.

        :param query: Either a :class:`str` that contains the query text,
              or a :class:`~.PreparedStatementInfo` that represents a pre-prepared query.
        """
        async with self.query(query, *params, **kwargs) as q:
            return await q.row_count()


class QueryResult:
    """
    Wraps the execution of a query. This can be asynchronously iterated over in order to get
    incoming data rows.
    """

    def __init__(self, iterator: AsyncIterator[PostgresMessage]):
        """
        :param iterator: An iterator of :class:`.PostgresMessage` instances returned from the
                         server.
        """
        self._iterator = iterator
        self._row_count = -1

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return False

    def __aiter__(self):
        return self

    async def __anext__(self):
        # infinitely loops until we get a message we care about

        while True:
            next_message = await self._iterator.__anext__()
            if isinstance(next_message, DataRow):
                return next_message
            elif isinstance(next_message, CommandComplete):
                # some messages don't have a row count, e.g. CREATE
                if next_message.row_count is None:
                    self._row_count = 0
                else:
                    self._row_count = next_message.row_count

                raise StopAsyncIteration

    async def row_count(self):
        """
        Gets the row count for this query.

        .. warning::

            This will discard any remaining data rows in the currently executing query.
        """
        if self._row_count >= 0:
            await anyio.sleep(0)  # checkpoint
            return self._row_count

        async for _ in self:
            pass

        return self._row_count


# noinspection PyProtectedMember
@asynccontextmanager
async def open_database_connection(
    address_or_path: Union[str, PathLike],
    username: str,
    *,
    port: int = 5432,
    password: str = None,
    database: str = None,
    ssl_context: SSLContext = None,
) -> AsyncContextManager[AsyncPostgresConnection]:
    """
    Opens a new connection to the PostgreSQL database server. This is an asynchronous context
    manager.

    .. code-block:: python3

        async with open_database_connection("localhost", username="postgres") as db:
            ...

    Required parameters:

    :param address_or_path: The address of the server or the *absolute path* of its Unix socket.
    :param username: The username to authenticate with.

    Optional parameters:

    :param port: The port to connect to. Ignored for unix sockets.
    :param password: The password to authenticate with.
    :param database: The database to connect to. Defaults to the username.
    :param ssl_context: The SSL context to use for TLS connection. Enables TLS if specified.
    """
    if address_or_path.startswith("/"):
        sock = await anyio.connect_unix(address_or_path)
    else:
        sock = await anyio.connect_tcp(remote_host=address_or_path, remote_port=port)

    try:
        if ssl_context:
            await sock.send(SSL_MESSAGE)
            response = await sock.receive(1)
            if not check_if_tls_accepted(response):
                raise ProtocolParseError("Requested TLS, but server said no")

            sock = await TLSStream.wrap(
                sock, hostname=address_or_path, ssl_context=ssl_context, standard_compatible=True
            )
    except BaseException:
        with anyio.move_on_after(delay=0.5, shield=True):
            await sock.aclose()

        raise

    async with sock:
        protocol = SansIOClient(username, database, password)
        conn = AsyncPostgresConnection(
            sock,
            protocol,
        )

        await conn._do_startup()
        await conn.wait_until_ready()

        # this sucks but we send a Terminate in the normal case, a Terminate in the case of a
        # database error, and a regular close in all other cases.
        try:
            yield conn
            await conn._terminate()
        except BaseDatabaseError:
            await conn._terminate()
            raise



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pg_purepy']

package_data = \
{'': ['*']}

install_requires = \
['anyio>=3.1.0,<4.0.0',
 'attrs>=21.2.0,<22.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'scramp>=1.4.0,<2.0.0']

extras_require = \
{':python_version < "3.10"': ['async_generator>=1.10,<2.0'],
 'docs': ['Sphinx>=3.0,<4.0',
          'sphinxcontrib-trio>=1.1.2,<2.0.0',
          'sphinx-autodoc-typehints>=1.12.0,<2.0.0',
          'sphinx-rtd-theme>=0.5.2,<0.6.0',
          'sphinx-inline-tabs>=2021.4.11-beta.9,<2022.0.0']}

setup_kwargs = {
    'name': 'pg-purepy',
    'version': '0.7.0',
    'description': 'A pure-Python anyio-based PostgreSQL adapter.',
    'long_description': 'pg-purepy\n=========\n\n.. image:: https://img.shields.io/pypi/pyversions/pg-purepy?style=flat-square\n    :alt: PyPI - Python Version\n\npg-purepy is a pure-Python PostgreSQL wrapper based on the `anyio`_ library.\n\nA lot of this library was inspired by the `pg8000`_ library. Credits to that.\n\nRead the docs at https://pg.py.veriny.tf/.\n\n.. _anyio: https://github.com/agronholm/anyio\n.. _pg8000: https://github.com/tlocke/pg8000',
    'author': 'Lura Skye',
    'author_email': 'l@veriny.tf',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
